#### Tutorial on Uncertainty Quantification in Machine Learning
####
####  Boston Consulting Group,  April 6/7, 2022

## The data throughout were downloaded from 
##
## https://www.kaggle.com/datasets/ruiqurm/lianjia
##
 
# Here we make use of the ranger and gbm libraries

library('ranger')
library('gbm')

# Loading in the Beijing Housing data

data = read.csv('BeijingRealEstate2.csv')

# Make sure categorical features are categorical
faccols = c(11,13,14,16:19)
for(i in 1:length(faccols)){ data[,faccols[i]] = as.factor(data[,faccols[i]]) }

# But also that categorical columns are categorical
data$constructionTime = as.numeric(data$constructionTime)
data$livingRoom = as.numeric(data$livingRoom)
data$drawingRoom = as.numeric(data$drawingRoom)
data$kitchen = as.numeric(data$kitchen)
data$bathRoom = as.numeric(data$bathRoom)


# Remove missing values
data = data[apply(is.na(data),1,sum)==0,]

# Re-order columns and remove an obvious outlier
p = ncol(data)
data0 = data[-99482,]

# Select 10 small sets of data at random, this will include both train and test data
set.seed('05032022')

ntrain = 3000  # Training set
ntest = 1500   # Testing set
nvalid = 500   # Validation will be taken as the end of the training set 
nsets = 10
ntree = 1000
index = matrix(sample(nrow(data0),(ntrain+ntest)*nsets,replace=FALSE),ntrain+ntest,nsets)


###### Prediction Intervals for Regression

# First train a model to predict total price
rmod = ranger(totalPrice~.,data=data0[index[1:ntrain,1],],keep.inbag=TRUE,num.trees=ntree)

# Now we'll get test-set predictions and errors
testpreds = predict(rmod,data0[index[ntrain+(1:ntest),1],])
testerr = data0[index[ntrain+(1:ntest),1],1] - testpreds$predictions

plot(testpreds$predictions,testerr,log='x')
abline(h=0)

# Calculate quantiles of erros based on

# Asymetric distribution
confquant1 = quantile(testerr[(nvalid+1):ntest],c(0.025,0.975))

# Out-of-bag errors
confquant2 = quantile( data0[index[1:ntrain,1],]$totalPrice - rmod$predictions, c(0.025,0.975) )

# Symetrized test errors
confquant3 = quantile(abs(testerr[(nvalid+1):ntest]),0.95)


# Histogram of test set errors
par(mar=c(5,5,2,2))
hist(testerr,100,col='blue',xlab='error',ylab='',cex.lab=1.5,cex.axis=1.5,main='')
abline(v=confquant1,col='darkblue',lwd=2)
abline(v=c(-1,1)*confquant3,col='cyan',lwd=2)


# This plot produces responses by fitted values along with error bands
plot(testpreds$predictions[1:nvalid],data0[index[ntrain+(1:nvalid)],1],col='darkblue',cex=2,
	xlab='f(X)',ylab='Y',cex.lab=1.5,cex.axis=1.5)
abline(c(0,1))
abline(c(confquant1[1],1),col='darkblue',lwd=2)
abline(c(confquant1[2],1),col='darkblue',lwd=2)
abline(c(confquant2[1],1),col='red',lwd=2)
abline(c(confquant2[2],1),col='red',lwd=2)
abline(c(confquant3,1),col='cyan',lwd=2)
abline(c(-confquant3,1),col='cyan',lwd=2)
legend('topleft',c('test set','symmetric','out of bag'),col=c('darkblue','cyan','red'),cex=1.5,lty=1,lwd=2)

# What does the coverage look like for data that have not been used to produce
# bands?

mean(  testerr[1:nvalid] > confquant1[1] & testerr[1:nvalid] < confquant1[2] )
mean(  testerr[1:nvalid] > confquant2[1] & testerr[1:nvalid] < confquant2[2] )
mean(  abs(testerr[1:nvalid]) <confquant3 )


## Non-constant error bands.  h

# These are the midpoints of predicted value bins that we will
# divide the data into:
pred.binmids = 100 + 200*(0:6)

# Let's plot observated response by fitted values again
plot( testpreds$predictions, data0[index[ntrain+(1:ntest)],1],xlim=c(0,1200),col='darkblue',cex=2,
	xlab='f(X)',ylab='Y',cex.lab=1.5,cex.axis=1.5)
abline(v=pred.binmids,col='cyan',lwd=2)
abline(c(0,1))

# This will calculate standard deviations in each bin
which.bin = as.factor( floor( testpreds$predictions/200) )
sds = tapply( testerr, which.bin, sd)
sdfun = approxfun(pred.binmids,sds[1:7])

# And we can add 1:1 line +/- 2SD lines to the plot
lines(pred.binmids,pred.binmids+2*sds[1:7],lwd=2,col='red')
lines(pred.binmids,pred.binmids-2*sds[1:7],lwd=2,col='red')



# Here we produce an explicit model for the standard deviation
# as a function of predictions. 

# In this case, we'll base it solely on bins we've calulated.
# Fitting squared errors tends to make the model susceptible to
# large outliers. 

sdlin = lm(sds[1:7]~I(pred.binmids^2))

plot(pred.binmids,sds[1:7],col='darkblue',cex=2,
	xlab='f(X)',ylab='sd(Y-F(X))',cex.lab=1.5,cex.axis=1.5)
lines(pred.binmids,sdlin$fitted.values,col='darkgreen',lwd=2)



# As an alternative, we can simply look at the residuals for the log 
# response as fitted by the log prediction. 

logtesterr = log10(data0[index[ntrain+(1:ntest),1],1]) - log10(testpreds$predictions)

# In this case, constant confidence bands are not so bad:

lconfquant1 = quantile(logtesterr[(nvalid+1):ntest],c(0.025,0.975))
lconfquant2 = quantile( log10(data0[index[1:ntrain,1],]$totalPrice) - log10(rmod$predictions), c(0.025,0.975) )
lconfquant3 = quantile( abs(logtesterr[(nvalid+1):ntest]),0.95)

plot(log10(testpreds$predictions[1:nvalid]),log10(data0[index[ntrain+(1:nvalid)],1]),col='darkblue',cex=2,
	xlab='log(f(X))',ylab='log(Y)',cex.lab=1.5,cex.axis=1.5)
abline(c(0,1))
abline(c(lconfquant1[1],1),col='darkblue',lwd=2)
abline(c(lconfquant1[2],1),col='darkblue',lwd=2)
abline(c(lconfquant3,1),col='cyan',lwd=2)
abline(c(-lconfquant3,1),col='cyan',lwd=2)



# If we re-rexponentiate, we can compare all the bands we have produced so far:

# Some points to plot modelled bands
fit.pts=seq(1.5,3.5,by=0.05)

plot(testpreds$predictions[1:nvalid],data0[index[ntrain+(1:nvalid)],1],col='darkblue',cex=2,
	xlab='f(X)',ylab='Y',cex.lab=1.5,cex.axis=1.5)
abline(c(0,1))
abline(c(confquant1[1],1),col='darkblue',lwd=2)
abline(c(confquant1[2],1),col='darkblue',lwd=2)
abline(c(confquant3,1),col='cyan',lwd=2)
abline(c(-confquant3,1),col='cyan',lwd=2)
lines(pred.binmids,pred.binmids+2*sds[1:7],lwd=2,col='green')
lines(pred.binmids,pred.binmids-2*sds[1:7],lwd=2,col='green')
lines(pred.binmids,pred.binmids+2*sdlin$fitted.values,col='darkgreen',lwd=2)
lines(pred.binmids,pred.binmids-2*sdlin$fitted.values,col='darkgreen',lwd=2)
lines(10^(fit.pts),10^(lconfquant1[1]+fit.pts),lwd=2,col='red')
lines(10^(fit.pts),10^(lconfquant1[2]+fit.pts),lwd=2,col='red')
legend('topleft',c('constant','symmetric','binned','quadratic','log'),
	col=c('darkblue','cyan','green','darkgreen','red'),cex=1.5,lty=1,lwd=2)

# In looking at uncertainty quantification below, we'll use log(totalPrice) as 
# the response because it is less sensitive to large values in the response variable. 



# Finally we will conformalize our +/- 2 bin sd standard errors 
# For this, we will use a second test set

testpreds2 = predict(rmod,data0[index[ntrain+(1:ntest),2],])
testerr2 = data0[index[ntrain+(1:ntest),2],1] - testpreds2$predictions
which.bin2 = as.factor( floor( testpreds2$predictions/200) )
id = 2*(1 + (abs(testerr2) < 2*sds[which.bin2]))

id = 2*(1 + (abs(testerr2) < 2*sdfun(testpreds2$predictions)))

uq = quantile( pmax(testerr2 - 2*sdfun(testpreds2$predictions),0),0.975,na.rm=TRUE)
lq = quantile( pmin(testerr2 + 2*sdfun(testpreds2$predictions),0),0.975,na.rm=TRUE)


# and work out how far outside the 2sd bounds they are

plot(testpreds2$predictions,testerr2,log='x',ylim=c(-1000,2000),col=id,cex=2,
	xlab='f(X)',ylab='Y',cex.lab=1.5,cex.axis=1.5)
abline(h = 0)
lines(pred.binmids,2*sds[1:7],lwd=2,col='darkblue')
lines(pred.binmids,-2*sds[1:7],lwd=2,col='darkblue')
lines(pred.binmids,2*sds[1:7]+uq,lwd=2,col='red',lty=2)
lines(pred.binmids,-2*sds[1:7]+lq,lwd=2,col='red',lty=2)
legend('topleft',c('2 SD','conformal update'),lwd=2,lty=c(1,2),col=c('darkblue','red'),cex=1.5)

##### Calibration for classification models

# Here we'll predict whether totalPrice is above 325, chosen as approximately
# the median value in the data

hist(log10(data0[index[1:ntrain,1],1]),100)
median(log10(data0[index[1:ntrain,1],1]))

data0$hprice = data0$totalPrice > 325

# Prediction model for random forests
rmod2 = ranger( hprice~.-totalPrice,data=data0[index[1:ntrain,1],],probability=TRUE)


# Test set fitted values
class.preds = predict(rmod2,data0[index[ntrain+(1:ntest)],])

# We'll put these into 20 bins by fitted value
prob.bins = as.factor( floor(class.preds$predictions[,1]*20) )

# And caluclate the proportion of 1's in each bin
qs = tapply( 1-as.numeric( data0[index[ntrain+(1:ntest)],]$hprice),prob.bins,sum)
ns = summary(prob.bins)


# Plot these by bin midpoint
prob.binmids = 0.025 + seq(0,0.95,by=0.05)

plot( prob.binmids, qs/ns,xlab='f(x)',ylab='Y',cex.lab=1.5,cex.axis=1.5,col='darkblue',cex=2)
points(class.preds$predictions[,2],1*data0[index[ntrain+(1:ntest)],]$hprice)
abline(c(0,1))
abline(v=seq(0,1,by=0.05),col='red')

# The relationship is more nearly linear on the logit scale
plot( qlogis(prob.binmids), qlogis(qs/ns),xlab='logit(f(x))',ylab='logit(p(x))',cex.lab=1.5,cex.axis=1.5,col='darkblue',cex=2)
abline(c(0,1))

# Here's a quick linear model based on values in the bins
plattlm = lm( qlogis(qs/ns)[1:18] ~ qlogis(prob.binmids)[1:18])
abline(plattlm,col='red',lwd=2)

# And add the calibrated predictions to the model. 
plot( prob.binmids, qs/ns,xlab='f(x)',ylab='Y',cex.lab=1.5,cex.axis=1.5,col='darkblue',cex=2)
points(class.preds$predictions[,2],1*data0[index[ntrain+(1:ntest)],]$hprice)
abline(c(0,1))
lines(prob.binmids[1:18], plogis(plattlm$fitted.values),col='red',lwd=2)


## It is also useful to see the effect of using an overfit version of 
## Adaboost.  For our purposes, we will use 1000 relatvely deep treees 

gb.classmod = gbm(as.numeric(hprice==1)~.-totalPrice,data=data0[index[1:ntrain,i],],
    n.trees=1000,interaction.depth=16,distribution="adaboost")

gb.classpreds = predict(gb.classmod, data0[index[ntrain+(1:ntest)],])

# If we plot predicted probabilities, these cluster near 0 or 1. 
par(mar=c(5,5,2,2))
hist(plogis(gb.classpreds),100,xlab='f(x)',ylab='',main='',cex.lab=1.5,cex.axis=1.5,col='blue')
title('Predicted Probabilities',cex.main=2)

# Dividing these up into bins we can see the overconfidence at the edges. 
# Note that predictions gives logit probabilities in gbm. 
gb.probbins = as.factor( floor(plogis(gb.classpreds)*20) )

gb.qs = tapply( data0[index[ntrain+(1:ntest)],]$hprice,gb.probbins,sum)
gb.ns = summary(gb.probbins)

plot( prob.binmids, gb.qs/gb.ns,xlab='f(x)',ylab='Y',cex.lab=1.5,cex.axis=1.5,col='darkblue',cex=2,ylim=c(0,1))
points(plogis(gb.classpreds),1*data0[index[ntrain+(1:ntest)],]$hprice)
abline(c(0,1))
abline(v=seq(0,1,by=0.05),col='cyan')

# Here's a straight glm on fitted values
gb.glm = glm( data0[index[ntrain+(1:ntest)],]$hprice ~ gb.classpreds, family='binomial')

points( plogis(gb.classpreds), gb.glm$fitted.values, pch='.',cex=3,col='red')

plot( qlogis(prob.binmids), qlogis(gb.qs/gb.ns),xlab='logit(f(X))',ylab='logit(p(X))',cex.lab=1.5,cex.axis=1.5,col='darkblue',cex=2)
abline(c(0,1))
abline(gb.glm,col='red',lwd=2)




##### Confidence Intervals

# Some parameters for bootstrapping
nboot = 20

# Choosing 5 points where we will make predictions
npred=5
testind = sample( (1:nrow(data))[-as.vector(index)], 5)

qpts = data[testind,]


# Set up storage arrays for 

# Predicted values for each data set, and each repeat of
# simple random numbers
qpreds = array(NA,c(nsets,nboot,npred,ntree))

# Predicted values for each data set and each separate 
# bootstrap of it. 
qpredsboot = array(NA,c(nsets,nboot,npred))

# For each tree, this records which observations were 
# available to it, and how many times they were replicated
inbags = array(NA,c(nsets,nboot,ntrain,ntree))

# These will record the Infinitesimal Jacknife estimates and
# the bias corrections for them. 
IJs = array(NA, c(nsets,nboot,npred))
IJcs = array(NA, c(nsets,nboot,npred))

# Loop over training sets, boostrap (or re-run random forests)
# for each one. 

for(i in 1:nsets){
	for(j in 1:nboot){
		# Random forest model
		tempmod = ranger(log(totalPrice)~.-hprice,data=data0[index[1:ntrain,i],],keep.inbag=TRUE,num.trees=ntree,sample.fraction=0.5)

		# Record predictions and inbag numbers
		inbags[i,j,,] = matrix( unlist(tempmod$inbag.counts), ntrain, ntree, byrow=FALSE)		
		preds = predict(tempmod,qpts,predict.all=TRUE)
		qpreds[i,j,,] = preds$predictions
		
		# Obtain IJ variance
		IJs[i,j,] = apply( cov(t(preds$predictions),t(inbags[i,j,,]))^2, 1, sum )
		
		# Bias correction for IJ
		IJcs[i,j,] = 0.5*ntrain/ntree*apply(preds$predictions,1,var)
		
		
		# Instead of just re-runing RF, here we do this on a bootstrap sample
		tempmod = ranger(log(totalPrice)~.-hprice,data=data0[sample(index[1:ntrain,i],replace=TRUE),],
			num.trees=ntree)
		preds = predict(tempmod,qpts,predict.all=FALSE)
		qpredsboot[i,j,] = preds$predictions
	}
	print(i)
}


# Now obtain predictions for each
qPred = apply(qpreds,1:3,mean)

# Plots of replicated models 
boxplot(qPred[1,,],xlab='X point',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,col='blue',ylim=c(5.2,6.8))

# Plots of values over each of 10 independent data sets
boxplot(qPred[,1,],xlab='X point',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,col='red',ylim=c(5.2,6.8))

# For first prediction point, variability in repreated runs for each data set
boxplot(t(qPred[,,1]),col='blue',xlab='Data Set',ylab='f(x)',cex.lab=1.5,cex.axis=1.5)

# For first prediction point, variability in bootstrap samples for each data set
boxplot(qpredsboot[1,,],xlab='X point',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,col='darkgreen',ylim=c(5.2,6.8))


# Ratio of MC to Sample variability
apply( apply(qPred,c(1,3),var),2,mean)/apply( apply(qPred,c(2,3),var),2,mean)



# Plot comparing replicates to bootstraps over the 10 data sets

# This just arranges columns in the right order
allvar = cbind(t(qPred[,,1]),t(qpredsboot[,,1]))[,rep(1:nsets,each=2)+rep(c(0,nsets),nsets)]

# And we then call boot
boxplot(allvar,col=rep( c('blue','darkgreen'),nsets),xaxt='n',cex.axis=1.5,xlab='Data Set',cex.lab=1.5)
axis(1,at=seq(1.5,19.5,by=2),labels=1:10,cex.axis=1.5)
legend('topright',c('Monte Carlo','Bootstrap'),lwd=2,lty=1,col= c('blue','darkgreen'),cex=1.5)


## Bootstrap Confidence intervals

# Standard deviation between bootstraps
bootsd = apply(qpredsboot,c(1,3),sd)

# +/- 2 Sd
bootupci = qPred[,1,] + 2*bootsd
bootdownci = qPred[,1,] - 2*bootsd

# Plot the 10 predictions, their average and the intervals. 
par(mar=c(5,5,2,2))
plot(rep(0,nsets), qPred[,1,1],pch=8,cex=3,col='darkblue',xaxt='n',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,
	xlab='Data Set',xlim=c(-0.5,10.5),ylim=c(min(bootdownci[,1]),max(bootupci[,1])))
for(i in 1:nsets){ lines( rep(i,2), c(bootupci[i,1],bootdownci[i,1]),lwd=3,col='darkgreen') }
abline(h = qPred[,1,1],col='darkblue',lty=3)
abline(h = mean(qPred[,1,1]),col='darkblue',lty=2,lwd=2)
axis(1,at=0.25+(1:10),label=1:10,cex.axis=1.5)






## Instead of confidence intervals, use reproduction intervals
## These are just sqrt(2) times as wide. 

bootupri = qPred[,1,] + 2*sqrt(2)*bootsd
bootdownri = qPred[,1,] - 2*sqrt(2)*bootsd

# We'll make sure that we threshold Infinitesimal Jacknife estimates at 0. Average
# 5 random forests to stabilize

IJc = apply( ((IJs - IJcs)*(IJs > IJcs))[,1:5,,drop=FALSE],c(1,3),mean)

# And here are reproduction intervals based on IJ

IJup = qPred[,1,] + 2*sqrt(2)*sqrt(IJc)
IJdown = qPred[,1,] - 2*sqrt(2)*sqrt(IJc)


# Same plot as confidence intervals above, but this time we should cover all 
# fitted values from each data set.  We include both Bootstrap and IJ intervals

plot(rep(0,nsets), qPred[,1,1],pch=8,cex=3,col='darkblue',xaxt='n',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,
	xlab='Data Set',xlim=c(-0.5,10.5),ylim=c(min(bootdowns[,1]),max(bootups[,1])))
for(i in 1:nsets){ lines( rep(i,2), c(bootupri[i,1],bootdownri[i,1]),lwd=3,col='darkgreen') }
abline(h = qPred[,1,1],col='darkblue',lty=3)
for(i in 1:nsets){ lines( rep(i+0.5,2), c(IJup[i,1],IJdown[i,1]),lwd=3,col='red') }
legend('bottomleft',c('boot','IJ'),lwd=3,lty=1,col=c('darkgreen','red'),cex=1.5)
axis(1,at=0.25+(1:10),label=1:10,cex.axis=1.5)


#### Extending the Infinitesimal Jacknife. 

# Here we apply it to an ensemble of boosted models

# Predictions over the 10 data sets for our predicted values
gb.pred = matrix(NA,nsets,npred)

# Predictions for each test set
gbm.testpred = matrix(NA,nsets,ntest)

# Places to record partial dependence plots over 'square'
spts = seq(50,200,by=1.5)
gbm.pdpall = matrix(NA,nsets,length(spts))


# For each data set obtain fitted values, we'll also look
# at over all test set error. 

for(j in 1:nsets){
	gbmod2 = gbm(log(totalPrice)~.-hprice,data=data0[index[1:ntrain,j],])
	gb.pred[j,] = predict(gbmod2, qpts)
	
	tempplot = plot(gbmod2,5,return.grid=TRUE)
	gbm.pdpall[j,] = approx(tempplot$square,tempplot$y,spts)$y
	
	gbm.testpred[j,] = predict(gbmod2, data0[index[ntrain+(1:ntest),j],])
}

## Now to build an ensemble, also for each data set

nmod = 500

# Record fitted values
gb.bootpreds = array(NA,c(nmod,nsets,npred))

# How many times a sample was included in each subsample
gb.inbag = array(0,c(nmod,ntrain))

# Values from partial dependence plots
gb.boot.pdp = array(NA,c(nmod,nsets,length(spts)))

# Test set fitted values
bootgb.testpred = array(NA,c(nmod,nsets,ntrain))

# Built 500 boosted models
for(b in 1:nmod){
	
	# Select indices (we'll take the same ones from each data set -- this is just
	# for convenience).
	ss = sample(1:ntrain,size=ntrain/2,replace=TRUE)
	
	# This calculates the inbag numbers for each bootstrap and
	# each data point
	ns = summary(as.factor(ss),maxsum=ntrain/2)
	gb.inbag[b, as.numeric(names(ns))] = ns
		
	for(j in 1:nsets){
		# Data points that we are using
		si = index[ss,j]
	
		tempmod = gbm(log(totalPrice)~.-hprice,data=data0[si,],distribution='gaussian',n.trees=100)
		gb.bootpreds[b,j,] = predict(tempmod,qpts)
		
		# Obtains partial dependence plots and registers them to the same points
		tempplot = plot(tempmod,5,return.grid=TRUE)
		gb.boot.pdp[b,j,] = approx(tempplot$square,tempplot$y,spts)$y
		
		# Predictions on the test set
		bootgb.testpred[b,j,] = predict(tempmod,data0[index[ntrain+(1:ntest),j],])
	}

}
		

# Let's calculate test set error for GBM and its ensemble version

GBM.TestErr = rep(NA,nsets)
eGBM.TestErr = rep(NA,nsets)
for(j in 1:nsets){
	GBM.TestErr[j] =
	   mean( (log(data0[index[ntrain+(1:ntest),j],]$totalPrice)-gbm.testpred[j,])^2 )
	eGBM.TestErr[j] =  
	   mean( (log(data0[index[ntrain+(1:ntest),j],]$totalPrice)-apply(bootgb.testpred[,j,],2,mean))^2 )
}		
		
# Reduction in error from ensembling is

1 - eGBM.TestErr/GBM.TestErr		
		
# Now let's look at bootstrap and IJ estimates for predictions

gb.bootsd = apply(gb.bootpreds,2:3,sd)

gb.bootup = gb.pred + 2*sqrt(2)*gb.bootsd
gb.bootdown = gb.pred - 2*sqrt(2)*gb.bootsd

# First, take the mean for the ensemble prediction
egb.pred = apply(gb.bootpreds,2:3,mean)

# This is the infinitesimal jacknife calculation
gb.IJ = matrix(NA,10,5)
for(i in 1:npred){
	gb.IJ[,i] = apply( cov(gb.inbag,gb.bootpreds[,,i])^2,2,sum) 		
}		
	
# IJ correction, we threshold at 0
gb.IJcs = 0.5*ntrain/nmod*gb.bootsd^2
gb.IJc = (gb.IJ-gb.IJcs)*(gb.IJ>gb.IJcs)

gb.IJup = egb.pred + 2*sqrt(2)*sqrt(gb.IJc)		
gb.IJdown = egb.pred - 2*sqrt(2)*sqrt(gb.IJc)	
	
# Here is fitted values for GBM on the first data set along with bootstrap
# confidence intervals, and for the ensemble GBM with IJ confidence intervals

plot(0.75+(0:4),gb.pred[1,],pch=8,cex=2,col='darkgreen',xaxt='n',xlab='X Point',ylab='f(x)',
		cex.lab=1.5,cex.axis=1.5,xlim=c(0.5,5.5),ylim=c(min(gb.bootdown[1,]),max(gb.bootup[1,])))
points(1.25+(0:4),egb.pred[1,],pch=8,cex=2,col='red')
for(i in 1:npred){
	lines(rep(0.75+i-1,2),c(gb.bootdown[1,i],gb.bootup[1,i]),lwd=3,col='darkgreen')
	lines(rep(1.25+i-1,2),c(gb.IJdown[1,i],gb.IJup[1,i]),lwd=3,col='red')
}
legend('bottomleft',c('boot','IJ'),lwd=3,lty=1,col=c('darkgreen','red'),cex=1.5)
axis(1,at=1:5,label=1:5,cex.axis=1.5)


# Here we will look at the first prediction point, but over the 10 data sets
# and give reproduction intervals calculated from each data set, too. 

plot(rep(0,nsets), gb.pred[,1],pch=8,cex=3,col='darkgreen',xaxt='n',ylab='f(x)',cex.lab=1.5,cex.axis=1.5,
	xlab='Data Set',xlim=c(-0.5,10.5),ylim=c(min(gb.bootdown[,1]),max(gb.bootup[,1])))
points(rep(0.5,nsets),egb.pred[,1],pch=8,col='red',cex=3)
for(i in 1:nsets){ lines( rep(i,2), c(gb.bootup[i,1],gb.bootdown[i,1]),lwd=3,col='darkgreen') }
abline(h = gb.pred[,1],col='darkgreen',lty=3)
for(i in 1:nsets){ lines( rep(i+0.5,2), c(gb.IJup[i,1],gb.IJdown[i,1]),lwd=3,col='red') }
abline(h = egb.pred[,1],col='red',lty=3)
legend('bottomleft',c('boot','IJ'),lwd=3,lty=1,col=c('darkgreen','red'),cex=1.5)
axis(1,at=0.25+(1:10),label=1:10,cex.axis=1.5)



# Here we will look at producing confidence intervals for partial dependence plots
# both for the first GBM model (via bootstrap), and for the ensemble GBM with IJ
# estimates.

# For this, we'll look at just the first data set. 

# Bootstrap SD of the PDP values
pdpbootsd = apply(gb.boot.pdp[,1,],2,sd)

# Inifinitesimal Jacknife
pdpIJ = apply( cov(gb.inbag,gb.boot.pdp[,1,])^2, 2, sum) - 0.5*ntrain/nmod*pdpbootsd^2

# And the ensemble PDP is the mean of the bootstraps
pdpmean = apply(gb.boot.pdp[,1,],2,mean)


matplot(spts,t(gb.boot.pdp[,1,]),type='l',col='lightblue',lty=1,xlab='square',ylab='Partial Dependence',
	cex.lab=1.5,cex.axis=1.5)
lines(spts,pdpmean,col='red',lwd=2)
lines(spts,pdpmean+2*sqrt(2*pdpIJ),col='red',lwd=2,lty=2)
lines(spts,pdpmean-2*sqrt(2*pdpIJ),col='red',lwd=2,lty=2)
lines(spts,gbm.pdpall[1,],col='darkgreen',lwd=2)
lines(spts,gbm.pdpall[1,]+2*sqrt(2)*pdpbootsd,col='darkgreen',lwd=2,lty=2)
lines(spts,gbm.pdpall[1,]-2*sqrt(2)*pdpbootsd,col='darkgreen',lwd=2,lty=2)
legend('bottomright',c('boot','IJ'),lwd=3,lty=1,col=c('darkgreen','red'),cex=1.5)


##### Combinations of models

# First lets look at boosting random forests.  First train one random forest,
# then train a second on the out-of-bag residuals from the first: 

# RF0 -- in-bag indicators, predictions, test set predictions

# Note that query point predictions are stored tree-by-tree, but
# we just use the average for test-set predictions since. 
brf.inbags0 = array(NA,c(nsets,ntrain,ntree))
brf.preds0 = array(NA,c(nsets,npred,ntree))
brf.testpreds0 = array(NA,c(nsets,ntest))

#RF1
brf.inbags1 = array(NA,c(nsets,ntrain,ntree))
brf.preds1 = array(NA,c(nsets,npred,ntree))
brf.testpreds1 = array(NA,c(nsets,ntest))

# Again, we will look at this for each data set, although we will only use the first
for(j in 1:nsets){
	tempdat = data0[index[1:ntrain,j],]
	
	# First random forest
	tempmod0 = ranger(log(totalPrice)~.-hprice,data=tempdat,keep.inbag=TRUE,num.trees=ntree,sample.fraction=0.5)
		
	brf.inbags0[j,,] = matrix( unlist(tempmod0$inbag.counts), ntrain, ntree, byrow=FALSE)
	
	brf.preds0[j,,] = predict(tempmod0,qpts,predict.all=TRUE)$predictions
	
	brf.testpreds0[j,] = predict(tempmod0,data0[index[ntrain+(1:ntest),j],])$predictions
	
	# Now replace the response with OOB residuals and train a new random forest
	tempdat$totalPrice = log(tempdat$totalPrice) - tempmod0$predictions 
	
	tempmod1 = ranger(totalPrice~.-hprice,data=tempdat,keep.inbag=TRUE,num.trees=ntree,sample.fraction=0.5)
	
	brf.inbags1[j,,] = matrix( unlist(tempmod1$inbag.counts), ntrain, ntree, byrow=FALSE)
	
	brf.preds1[j,,] = predict(tempmod1,qpts,predict.all=TRUE)$predictions
	
	brf.testpreds1[j,] = predict(tempmod1,data0[index[ntrain+(1:ntest),j],])$predictions

}

# Let's look at test set error for each of the 10 data sets
RF0.TestErr = rep(NA,nsets)   # First forest error
RF1.TestErr = rep(NA,nsets)   # Error of the sum of forests

for(j in 1:nsets){
	RF0.TestErr[j] = mean( (log(data0[index[ntrain+(1:ntest),j],]$totalPrice)-brf.testpreds0[j,])^2 )
	RF1.TestErr[j] = mean( (log(data0[index[ntrain+(1:ntest),j],]$totalPrice)-brf.testpreds0[j,]-brf.testpreds1[j,])^2 )
}

# Error reduction is
1 - RF1.TestErr/RF0.TestErr


# To produce confidence intervals, we need to obtain covariances
# between inbag indicators, and the predictions. We'll do this
# for the two random forests, as well as the ensemble of GBM's

bcov0 = cov( t(brf.inbags0[1,,]), t(brf.preds0[1,,]) )
bcov1 = cov( t(brf.inbags1[1,,]), t(brf.preds1[1,,]) )
gcov = cov(gb.inbag,gb.bootpreds[,1,])


# Infinitesimal jacknife calculations square these and sum, but we also 
# have to subtract bias corrections

brf0.IJ = apply( (bcov0)^2,2,sum) - 0.5*ntree/ntrain*apply(brf.preds0[1,,],1,var)
brf1.IJ = apply( (bcov0+bcov1)^2,2,sum) - 0.5*ntree/ntrain*apply(brf.preds0[1,,],1,var) - 0.5*ntree/ntrain*apply(brf.preds1[1,,],1,var)

# We need to average out the trees in the forsts for the individual trees

brf.mean.preds0 = apply(brf.preds0[1,,],1,mean)
brf.mean.preds1 = apply(brf.preds1[1,,],1,mean)

# Reproduction intervals are
brf0.down = brf.mean.preds0 - 2*sqrt(2*brf0.IJ)
brf0.up = brf.mean.preds0 + 2*sqrt(2*brf0.IJ)

brf1.down = brf.mean.preds0+brf.mean.preds1 - 2*sqrt(2*brf1.IJ)
brf1.up = brf.mean.preds0+brf.mean.preds1 + 2*sqrt(2*brf1.IJ)


# Getting the total range of reproduction intervals:
yrange = range( c(brf0.down,brf0.up,brf1.down,brf1.up) )

plot(0.75+(0:4),brf.mean.preds0,pch=8,cex=2,col='darkblue',xaxt='n',xlab='X Point',ylab='f(x)',
		cex.lab=1.5,cex.axis=1.5,xlim=c(0.5,5.5),ylim=yrange)
points(1.25+(0:4),brf.mean.preds0+brf.mean.preds1,pch=8,cex=2,col='red')
for(i in 1:npred){
	lines(rep(0.75+i-1,2),c(brf0.down[i],brf0.up[i]),lwd=3,col='darkblue')
	lines(rep(1.25+i-1,2),c(brf1.down[i],brf1.up[i]),lwd=3,col='red')
	lines(c(0.75,1.25)+i-1, brf.mean.preds0[i]+c(0,brf.mean.preds1[i]),col=2)
}
axis(1,at=1:5,label=1:5,cex.axis=1.5)
legend('bottomleft',c('RF0','RF0+RF1'),col=c('darkblue','red'),lwd=2,lty=1,cex=1.5)


# Finally,let's look at confidence intervals for the difference between 
# the ensemble GBM and the first random forest

# Here is the IJ calculation
diffIJ = apply( (bcov0-gcov)^2,2,sum) - 0.5*ntree/ntrain*apply(brf.preds0[1,,],1,var) - 0.5*ntrain/nmod*gb.bootsd[1,]^2

# The difference in points
diff.preds = egb.pred[1,] - brf.mean.preds0

# Reproduction intervals
diff.down = diff.preds - 2*sqrt(2*diffIJ)
diff.up = diff.preds + 2*sqrt(2*diffIJ)

# And lets produce a plot
plot(1:5,diff.preds,pch=8,col='darkblue',cex=2,xlab='X Point',ylab='GBM(X) - RF(X)',
	cex.lab=1.5,cex.axis=1.5,ylim=range(c(diff.down,diff.up)))
for(i in 1:npred){ lines( rep(i,2), c(diff.down[i],diff.up[i]),lwd=2,col='red') }
abline(h = 0)




#### Tests of variable importance 

# Here we'll conduct a conditional randomization test for renovation
#
# First we need a model for renovation for which we'll use ranger. 
# Note that this is coded as factorial so we will get a multinomial
# probability model. 

# We can do this entirely using the training data. 

renmod = ranger(renovationCondition~.-totalPrice,data=data0[index[1:ntrain,1],],probability=TRUE)

# We'll generate 500 replicates of simulated respones from these

newren = matrix(NA,ntrain,500)

for(i in 1:ntrain){
	newren[i,] = as.vector(t(rmultinom(500,1,renmod$predictions[i,]))%*%1:4)
}

# Now for each simulated renovation condition, we'll use it to replace
# renovation in the training set and examine the OOB r-squared of the
# resulting rannger model

nullrsquare = rep(NA,500)


for(s in 1:500){
	tempdat = data0[index[1:ntrain,1],]
	tempdat$renovationCondition = newren[,s]
	tempdat$renovationCondition = as.factor(tempdat$renovationCondition)
	tempmod = ranger(totalPrice~.,data=tempdat)
	nullrsquare[s] = tempmod$r.squared
}

# A p-value is then given by the proportion of null r-squares that 
# exceed what we saw for the original data

pval = mean(nullrsquare>rmod$r.square)
hist(nullrsquare,cex.lab=2,ylab='',xlab='R Square',
		main=substitute(paste(P(tilde(R)^2 > R^2),' = ',pval),list(pval=pval)),cex.main=2,col='blue') 
abline(v = rmod$r.square,col='red',lwd=2)


#### Causal inference for enovation. 

# Unlike the conditional randomiztion test, we will ultimately 
# use a regression model for which we care about distributional 
# assumptions so using log price as a response makces sense. 

# Here we have a model for log price on everything except renovation condition.  

rmod2 = ranger(log(totalPrice) ~ .-renovationCondition,data=data0[index[1:ntrain,1],])

# And we'll obtain test set predictions and errors. 

testpred2 = predict(rmod2,data0[index[ntrain+(1:ntest),1],])$predictions
testerr2 = log(data0[index[ntrain+(1:ntest),1],1]) - testpred2


# We also need to model renovation condition and obtain its
# test set predictions and errors

renoTestResp = model.matrix(~renovationCondition,data=data0[index[ntrain+(1:ntest),1],])
renoTestPred = predict(renmod,data0[index[ntrain+(1:ntest),1],])$predictions
renoTestErr = renoTestResp - renoTestPred

# Then we just need to establish if there is a relationship between the two:

par(mfrow=c(2,2),mar=c(5,5,1,1))
for(i in 1:4){ 
	plot(renoTestPred[,i],testpred2,ylab='Price Prediction',
		xlab=paste('Condition ',i,' Prediction'),cex.lab=2,cex.axis=1.5,col=4) 
	
}

summary(lm(testerr2~renoTestErr-1))
